import sys
from collections import namedtuple
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import streamlit as st
from negmas.helpers import distribute_integer_randomly
from negmas.helpers.inout import load, np
from negmas.inout import Scenario
from negmas.plots.util import TraceElement, plot_offline_run

from anl import DEFAULT_TOURNAMENT_PATH


def is_tournament_folder(base: Path):
    if not base.is_dir():
        return False
    for mark in ("negotiations", "results", "scenarios"):
        path = base / mark
        if not (path.exists() and path.is_dir()):
            return False
    for mark in ("scores.csv", "type_scores.csv", "details.csv", "all_scores.csv"):
        path = base / mark
        if not (path.exists() and path.is_file()):
            return False
    return True


def tournaments(base: Path = DEFAULT_TOURNAMENT_PATH):
    if is_tournament_folder(base):
        return base.absolute().parent, [base.name]
    return (
        base.absolute(),
        sorted(
            [
                str(_.relative_to(base))
                for _ in base.glob("**/*")
                if is_tournament_folder(_)
            ]
        )[::-1],
    )


@st.cache_data
def negotiations(path: Path):
    base = path / "negotiations"
    return sorted(
        [_.stem for _ in base.glob("*") if _.is_file() and _.suffix.endswith("csv")]
    )


@st.cache_data
def read_scenarios(path: Path):
    base = path / "scenarios"
    return sorted([_.name for _ in base.glob("*") if _.is_dir()])


ScoreDataFrames = namedtuple(
    "ScoresDataFrame", ["final_scores", "type_scores", "all_scores", "details"]
)


@st.cache_resource
def read_score_dfs(
    base: Path,
    tournament: str,
    selected_scenarios: list[str] | None,
    selected_partners: list[str] | None,
    selected_strategies: list[str] | None,
    read_type: bool = True,
    read_all: bool = True,
    metric: str = "advantage",
    stat: str = "mean",
):
    if selected_strategies or selected_partners or selected_scenarios:
        read_all = True
    scores = pd.read_csv(base / tournament / "scores.csv", index_col=0)
    type_scores = (
        pd.read_csv(base / tournament / "type_scores.csv", index_col=0, header=[0, 1])
        if read_type
        else None
    )
    all_scores = pd.read_csv(base / tournament / "all_scores.csv") if read_all else None
    details = pd.read_csv(base / tournament / "details.csv")
    if selected_partners is not None:
        cond = details.negotiator_types.astype(bool)
        for p in selected_partners:
            cond &= details.negotiator_types.str.contains(p)
        details = details.loc[cond, :]
        assert all_scores is not None
        all_scores = all_scores.loc[all_scores.partners.isin(set(selected_partners)), :]
    if selected_strategies is not None:
        assert all_scores is not None
        cond = details.negotiator_types.astype(bool)
        for p in selected_strategies:
            cond &= details.negotiator_types.str.contains(p)
        details = details.loc[cond, :]
        all_scores = all_scores.loc[
            all_scores.strategy.isin(set(selected_strategies)), :
        ]
    if selected_scenarios is not None:
        assert all_scores is not None
        s = set(selected_scenarios)
        details = details.loc[(details.scenario.isin(s)), :]
        all_scores = all_scores.loc[
            all_scores.scenario.isin(set(selected_scenarios)), :
        ]
    if selected_scenarios or selected_partners or selected_strategies:
        assert all_scores is not None
        type_scores = all_scores.groupby(["strategy"]).describe()
        scores = type_scores[(metric, stat)]
    return ScoreDataFrames(scores, type_scores, all_scores, details)


def read_scenarios_with_steps(
    scenarios: list[str], details: pd.DataFrame
) -> dict[str, int]:
    n_scenarios_found = len(details.scenario.unique())
    n_scenarios = len(scenarios)
    assert (
        n_scenarios_found == n_scenarios
    ), f"Found {n_scenarios_found} scenarios in details but we should only find {n_scenarios}"
    min_n_steps = details.groupby(["scenario"])["n_steps"].min().rename("n_steps_min")  # type: ignore
    max_n_steps = details.groupby(["scenario"])["n_steps"].max().rename("n_steps_max")  # type: ignore
    all_n_steps = pd.concat(
        (min_n_steps, max_n_steps), axis=1, ignore_index=False, join="outer"
    )
    assert np.all(all_n_steps["n_steps_min"] == all_n_steps["n_steps_max"])
    return min_n_steps.to_dict()


@st.cache_resource
def make_scenario_stats(
    scenarios: list[str],
    details: pd.DataFrame,
    grouper: str | list[str] = "scenario",
    fields=(
        "n_steps",
        "time_limit",
        "n_repetitions",
        "step",
        "time",
        "relative_time",
        "broken",
        "timedout",
        "agreement_rate",
        "reserved_value(First)",
        "reserved_value(Second)",
        "execution_time",
        "pareto_optimality",
        "nash_optimality",
        "kalai_optimality",
        "modified_kalai_optimality",
        "max_welfare_optimality",
        "has_error",
    ),
) -> pd.DataFrame:
    if "DoneConversion" not in details.columns:
        for x in ("negotiator_types", "negotiator_ids", "negotiator_names"):
            if x in grouper:
                lst = [
                    ",".join(z.split(".")[-1] for z in eval(_))
                    for _ in details[x].to_list()
                ]
                details[x] = lst
        details["DoneConversion"] = True
    n_scenarios_found = len(details.scenario.unique())
    n_scenarios = len(scenarios)
    assert (
        n_scenarios_found == n_scenarios
    ), f"Found {n_scenarios_found} scenarios in details but we should only find {n_scenarios}"
    if isinstance(grouper, str):
        grouper = [grouper]  # type: ignore
    cols = grouper + list(fields)  # type: ignore
    details["agreement_rate"] = details.agreement.astype(str) != "nan"
    details["r"] = details.reserved_values.apply(lambda x: eval(x))
    details[["reserved_value(First)", "reserved_value(Second)"]] = pd.DataFrame(
        details["r"].tolist(), index=details.index
    )
    details = details[cols]  # type: ignore
    return details.groupby(grouper).mean()  # type: ignore


metrics = [
    "advantage",
    "kalai_optimality",
    "max_welfare_optimality",
    "nash_optimality",
    "pareto_optimality",
    "partner_welfare",
    "time",
    "utility",
    "welfare",
]


def run(base: Path = DEFAULT_TOURNAMENT_PATH):
    st.set_page_config(
        page_title="your_title", layout="wide", initial_sidebar_state="auto"
    )
    st.write("### ANL Tournament Visualizer")
    base, t = tournaments(base)
    tournament = st.sidebar.selectbox("Tournament", t)
    if tournament is None:
        return
    st.write(f"#### Tournament: {tournament}")
    scenarios = read_scenarios(base / tournament)
    dfs = read_score_dfs(
        base,
        tournament,
        read_type=False,
        read_all=False,
        selected_partners=None,
        selected_scenarios=None,
        selected_strategies=None,
    )
    scores = dfs.final_scores
    steps_of = read_scenarios_with_steps(scenarios, dfs.details)
    # scenarios_of = dict(zip(steps_of.values(), steps_of.keys()))
    if st.sidebar.checkbox("Show Final Scores", value=True):
        st.write("**Final Scores**")
        st.dataframe(scores)
    with st.expander("Scenario Filter"):
        scenario_prefix = st.text_input("Scenario Prefix")
        selected_scenarios = st.multiselect("Specific Scenarios", scenarios)
        if not selected_scenarios:
            selected_scenarios = scenarios
        st.write("N. Steps")
        cola1, cola2 = st.columns(2)
        with cola1:
            selected_n_steps_min = st.number_input(
                label="min", min_value=1, key="min_n_steps", value=1
            )
        with cola2:
            selected_n_steps_max = st.number_input(
                label="max",
                min_value=selected_n_steps_min,
                value=1000_000,
                key="max_n_steps",
            )
        if selected_n_steps_max < selected_n_steps_min:
            st.write(
                "The maximum number of steps {selected_n_steps_max} is smaller than the minimum number {selected_n_steps_min}"
            )
            return
    with st.expander("Strategy Filter"):
        types = sorted(scores["strategy"].unique().tolist())
        type_ = st.multiselect("Strategies", types)
        partners = st.multiselect("Partners", types)
    prefix_selected = []
    if scenario_prefix:
        prefix_selected = [_ for _ in scenarios if _.startswith(scenario_prefix)]
    selected_scenarios = sorted(
        list(set(selected_scenarios).union(set(prefix_selected)))
    )
    # st.write(selected_scenarios)
    selected_scenarios = [
        _
        for _ in selected_scenarios
        if selected_n_steps_min <= steps_of[_] <= selected_n_steps_max
    ]
    st.write(
        f"Selected scenarios: {selected_scenarios if selected_scenarios  and len(selected_scenarios) < len(scenarios) else 'ALL'} and "
        f"Selected strategies: {type_ if type_ else 'ALL'} and "
        f"Selected partners: {partners if partners else 'ALL'}"
    )
    dfs = read_score_dfs(
        base,
        tournament,
        read_type=True,
        read_all=True,
        selected_partners=partners if partners else None,
        selected_scenarios=selected_scenarios if selected_scenarios else None,
        selected_strategies=type_ if type_ else None,
    )

    if st.sidebar.checkbox("Show Scenario Stats"):
        st.dataframe(make_scenario_stats(selected_scenarios, dfs.details))
    if st.sidebar.checkbox("Show Scenario x Strategy Stats"):
        st.dataframe(
            make_scenario_stats(
                selected_scenarios,
                dfs.details,
                grouper=["scenario", "negotiator_types"],
            )
        )
    if st.sidebar.checkbox("Show Cases Never Succeeding"):
        data = make_scenario_stats(
            selected_scenarios,
            dfs.details,
            grouper=["scenario", "negotiator_types"],
        )
        st.dataframe(data.loc[data.agreement_rate == 0, :])

    scores = dfs.final_scores
    st.write(f"Advantage (Score) for the selected set of scenarios and partners")
    st.dataframe(dfs.all_scores.groupby("strategy")["advantage"].describe())
    if st.sidebar.checkbox("Show Type Scores", value=False):
        st.write("**Type Scores**")
        metric = st.sidebar.multiselect("Metrics", metrics, default=("advantage",))
        df = dfs.type_scores
        if metric:
            df = df[[_ for _ in df.columns if _[0] in metric]]
        st.dataframe(df)

    if st.sidebar.checkbox("Show All Scores", value=False):
        st.write("**All Scores**")
        df = dfs.all_scores
        if selected_scenarios:
            df = df.loc[df.scenario.isin(selected_scenarios), :]
        if type_:
            df = df.loc[df.strategy.isin(type_), :]
        if partners:
            df = df.loc[df.partners.isin(partners), :]
        st.dataframe(df)
    all_negotiations = negotiations(base / tournament)
    if selected_scenarios:
        all_negotiations = [
            _
            for _ in all_negotiations
            if any(_.startswith(x) for x in selected_scenarios)
        ]
    if type_:
        all_negotiations = [_ for _ in all_negotiations if any(x in _ for x in type_)]

    if partners:
        all_negotiations = [
            _ for _ in all_negotiations if any(x in _ for x in partners)
        ]

    def display_negotiation(negotiation_set, neg_limits, indx=0, n_cols=1):
        n_negs = len(negotiation_set)
        for k in range(neg_limits[0], neg_limits[1]):
            negotiation = negotiation_set[k]
            st.write(negotiation)
            neg_path = base / tournament / "negotiations" / (negotiation + ".csv")
            results_path = base / tournament / "results" / (negotiation + ".json")
            parts = negotiation.split("_")
            scenario_name = parts[0]
            scenario = Scenario.load(base / tournament / "scenarios" / scenario_name)
            if scenario is None:
                return
            col1, col2 = st.columns(2)
            show_2d = show_offers = simple_offers = False
            x_var_offers = "relative_time"

            with col1:
                side_by_side = st.checkbox(
                    "Side-by-Side",
                    value=n_cols == 1 and n_negs > 1,
                    key=f"side_by_side_{indx}_{k}",
                )
            with col2:
                show_details = st.checkbox(
                    "Outcome Details", value=False, key=f"results_{indx}_{k}"
                )
            if side_by_side:
                pass
            else:
                col1, col2, col3, col4 = st.columns(4)
                with col1:
                    show_2d = st.checkbox(
                        "2D Plot",
                        value=True,
                        key=f"twod_{indx}_{k}",
                    )
                with col2:
                    show_offers = st.checkbox(
                        "Offers",
                        value=n_cols == 1,
                        key=f"offer_{indx}_{k}",
                    )
                with col3:
                    if show_offers:
                        simple_offers = st.checkbox(
                            "Simple",
                            value=False,
                            key=f"simple_{indx}_{k}",
                        )
                with col4:
                    if show_offers:
                        x_var_offers = st.selectbox(
                            "X-variable",
                            options=("relative_time", "time", "step"),
                            index=0,
                            key=f"xvar_offers_{indx}_{k}",
                        )

            # first_type = "_".join(parts[1:3])
            # second_type = "_".join(parts[3:5])
            trace_df = pd.read_csv(neg_path, index_col=None)
            trace = [
                TraceElement(
                    _[0],
                    _[1],
                    _[2],
                    _[3],
                    eval(_[4]),
                    {},
                    _[6],
                )
                for _ in trace_df.itertuples(index=False)
            ]
            results = load(results_path)
            names = [_.split(".")[-1] for _ in results["negotiator_names"]]
            if side_by_side:
                plot_offline_run(
                    trace,
                    ids=results["negotiator_ids"],
                    names=names,
                    ufuns=scenario.ufuns,  # type: ignore
                    agreement=results["agreement"],
                    broken=results["broken"],
                    has_error=results["has_error"],
                    timedout=results["timedout"],
                    ylimits=(0, 1),
                )
                st.pyplot(plt.gcf())
            else:
                if show_2d:
                    plot_offline_run(
                        trace,
                        ids=results["negotiator_ids"],
                        names=names,
                        ufuns=scenario.ufuns,  # type: ignore
                        agreement=results["agreement"],
                        broken=results["broken"],
                        has_error=results["has_error"],
                        timedout=results["timedout"],
                        only2d=True,
                        ylimits=(0, 1),
                    )
                    st.pyplot(plt.gcf())
                if show_offers:
                    plot_offline_run(
                        trace,
                        ids=results["negotiator_ids"],
                        names=names,
                        ufuns=scenario.ufuns,  # type: ignore
                        agreement=results["agreement"],
                        broken=results["broken"],
                        has_error=results["has_error"],
                        timedout=results["timedout"],
                        no2d=True,
                        simple_offers_view=simple_offers,
                        ylimits=(0, 1),
                        xdim=x_var_offers if x_var_offers else "relative_time",
                    )
                    st.pyplot(plt.gcf())

            if show_details:
                st.write("Results for this Negotiation:")
                st.write(results)
        st.markdown("""---""")

    if st.sidebar.checkbox("Show Negotiation", value=True):
        negotiation_set = st.sidebar.multiselect("Negotiation", all_negotiations)
        n_cols = st.sidebar.slider("N. Columns", 1, 4, value=1)
        negs_for_col = distribute_integer_randomly(
            len(negotiation_set), n_cols, min_per_bin=None  # type: ignore
        )
        negs_for_col = [_ for _ in negs_for_col if _]
        n_cols = len(negs_for_col)
        for i in range(1, n_cols):
            negs_for_col[i] += negs_for_col[i - 1]
        if n_cols:
            st.write(f"##### Negotiations")
            if n_cols == 1:
                display_negotiation(
                    negotiation_set, (0, len(negotiation_set)), n_cols=1
                )
            else:
                negs_for_col_limits = [(0, negs_for_col[0])] + list(
                    zip(negs_for_col[:-1], negs_for_col[1:])
                )
                cols = st.columns(n_cols)
                for indx, (col, neg_limits) in enumerate(
                    zip(cols, negs_for_col_limits)
                ):
                    with col:
                        if not neg_limits:
                            continue
                        display_negotiation(
                            negotiation_set, neg_limits, indx, n_cols=n_cols
                        )
    if st.sidebar.checkbox("Plot Stats", value=False):
        metric = st.sidebar.multiselect("Metric", metrics, default=("advantage",))
        df = pd.read_csv(base / tournament / "all_scores.csv")
        if selected_scenarios:
            df = df.loc[df.scenario.isin(selected_scenarios), :]
        if type_:
            df = df.loc[df.strategy.isin(type_), :]
        if partners:
            df = df.loc[df.partners.isin(partners), :]
        # st.dataframe(df)
        for m in metric:
            col1, col2, col3 = st.columns(3)

            with col1:
                show_table = st.checkbox(
                    "Show Table",
                    key=f"table_{m}",
                    value=True,
                )
                show_figs = st.checkbox(
                    "Show Graph",
                    key=f"graph_{m}",
                    value=True,
                )
            with col2:
                xvar = st.multiselect(
                    "X Variable",
                    ("strategy", "scenario", "partners"),
                    key=f"xvar_{m}",
                    default=("strategy",),
                )
            if len(xvar) < 1:
                st.write(f"Choose some variable to analyze")
            # elif len(xvar) > 2:
            #     st.write(
            #         f"You selected {len(xvar)} variables for analysis. We only support up to two."
            #     )
            else:
                if show_figs:
                    with col3:
                        graph = st.selectbox(
                            "Graph Type",
                            ("violin", "box", "KDE", "histogram"),
                            key=f"graph_type_{m}",
                        )
                    plt.figure()

                    if len(xvar) > 2:
                        newcol = "(" + ", ".join(xvar[1:]) + ")"
                        df[newcol] = df[xvar[1]]
                        for i in range(2, len(xvar)):
                            df[newcol] += ", " + df[xvar[i]]
                    elif len(xvar) == 2:
                        newcol = xvar[1]
                    else:
                        newcol = None
                    if graph == "violin":
                        sns.violinplot(df, x=xvar[0], y=m, hue=newcol)
                    if graph == "box":
                        sns.boxplot(
                            df,
                            x=xvar[0],
                            y=m,
                            hue=newcol,
                            notch=True,
                        )
                    elif graph == "KDE":
                        df.groupby(xvar)[m].plot(kind="kde", ylabel=m)
                        plt.legend()
                    elif graph == "histogram":
                        df.groupby(xvar)[m].plot(kind="hist", ylabel=m)
                        plt.legend()
                    plt.ylabel(m)
                    if len(xvar) > 1 or (
                        len(xvar) == 1 and len(df[xvar[0]].unique()) > 6
                    ):
                        plt.xticks(rotation=90)
                    st.pyplot(plt.gcf())
                    plt.close()
                if show_table:
                    st.dataframe(df.groupby(xvar)[m].describe())
    if st.sidebar.checkbox("Negotiation Distributions", value=False):
        vars = [
            "n_steps",
            "time_limit",
            "step",
            "has_error",
            "relative_time",
            "time",
            "n_repetitions",
        ]
        # index,running,waiting,started,step,time,relative_time,broken,timedout,agreement,results,n_negotiators,has_error,error_details,threads,last_thread,current_offer,current_proposer,current_proposer_agent,
        # n_acceptances,new_offers,new_offerer_agents,last_negotiator,utilities,max_utils,reserved_values,partners,params,scenario,run_id,execution_time,negotiator_names,negotiator_ids,
        # negotiator_types,negotiator_times,n_steps,time_limit,pend,pend_per_second,step_time_limit,negotiator_time_limit,annotation,mechanism_name,mechanism_type,effective_scenario_name,rep,
        # n_repetitions,pareto_optimality,nash_optimality,kalai_optimality,modified_kalai_optimality,max_welfare_optimality,__weakref__
        scenario_vars = st.sidebar.multiselect(
            "Variable", vars, default=("n_steps",), key="dist_key"
        )
        df = dfs.details
        for v in scenario_vars:
            plt.figure()
            df[v].plot(kind="hist")
            st.pyplot(plt.gcf())
            plt.close()


if __name__ == "__main__":
    run(Path(sys.argv[1]) if len(sys.argv) > 1 else DEFAULT_TOURNAMENT_PATH)



from negmas import MappingUtilityFunction
from negmas.inout import SAOMechanism
from negmas.outcomes import make_issue, make_os
from negmas.preferences.generators import make_zero_sum_pareto
from negmas.tournaments.neg.simple import SimpleTournamentResults

from anl.anl2024.negotiators.builtins import MiCRO
from anl.anl2024.runner import anl2024_tournament


def test_micro_agrees_with_itself():
    n_outcomes = 1000
    os = make_os([make_issue(n_outcomes)])
    ufun_values = make_zero_sum_pareto(n_outcomes)
    ufuns = [
        MappingUtilityFunction(
            dict(zip(os.enumerate_or_sample(), [_[i] for _ in ufun_values])),
            outcome_space=os,
            reserved_value=0.05 * i,
        )
        for i in range(2)
    ]
    m = SAOMechanism(n_steps=n_outcomes, outcome_space=os)
    for i in range(2):
        m.add(MiCRO(id=f"{i}", name=f"M{i}"), ufun=ufuns[i])
    m.run()
    assert m.agreement is not None


def test_micro_always_agrees_against_itself():
    results = anl2024_tournament(
        n_scenarios=50,
        competitors=[MiCRO],
        n_repetitions=1,
        nologs=True,
    )
    scores = results.scores

    failed_scenarios = scores.loc[
        (scores.strategy == "MiCRO")
        & (scores.partners == "MiCRO")
        & (scores.advantage < 0),
        "scenario",
    ].unique()
    assert not (failed_scenarios), f"Negative advanate: {failed_scenarios=}"
    failed_scenarios = scores.loc[
        (scores.strategy == "MiCRO")
        & (scores.partners == "MiCRO")
        & (scores.advantage < 1e-3),
        "scenario",
    ].unique()
    assert not (failed_scenarios), f"Failure to get agreement: {failed_scenarios=}"



#!/usr/bin/env python
import sys
from functools import partial
from pathlib import Path

import click
from streamlit.web import cli as stcli

from anl import DEFAULT_TOURNAMENT_PATH

click.option = partial(click.option, show_default=True)


@click.group()
def main():
    pass


@main.command(help="Opens the visualizer")
@click.option(
    "-f",
    "--folder",
    type=click.Path(file_okay=False, dir_okay=True),
    default=None,
    help=f"Folder containing logs of a tournament to open. If not given, All runs at {DEFAULT_TOURNAMENT_PATH} will be used",
)
@click.option(
    "-p",
    "--port",
    type=int,
    default=8501,
    help=f"The port to run the visualizer on",
)
@click.option(
    "-a",
    "--address",
    type=str,
    default="0.0.0.0",
    help=f"The address to run the visualizer on",
)
def show(folder: Path, port: int, address: str):
    # folder = Path(folder) if folder is not None else None
    if folder:
        sys.argv = [
            "streamlit",
            "run",
            str(Path(__file__).parent / "app.py"),
            str(folder),
        ]
    else:
        sys.argv = ["streamlit", "run", str(Path(__file__).parent / "app.py")]
    if port:
        sys.argv += ["--server.port", str(port)]
    if address not in ("", "default"):
        sys.argv += ["--server.address", address]
    sys.exit(stcli.main())


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""The ANL universal command line tool"""
import math
import sys
import warnings
from collections import defaultdict
from functools import partial
from itertools import chain
from pathlib import Path
from time import perf_counter
from typing import Iterable, List

import click
import click_config_file
import matplotlib.pyplot as plt
import negmas
from negmas.helpers import humanize_time, unique_name
from negmas.helpers.inout import load
from negmas.helpers.types import get_class
from negmas.inout import Scenario
from negmas.plots.util import plot_offline_run
from rich import print

import anl
from anl import DEFAULT_AN2024_COMPETITORS
from anl.anl2024.runner import (
    DEFAULT2024SETTINGS,
    DEFAULT_TOURNAMENT_PATH,
    GENERAROR_MAP,
    anl2024_tournament,
)

n_completed = 0
n_total = 0


def default_log_path():
    """Default location for all logs"""
    return Path.home() / "negmas" / "anl"


DB_FOLDER = default_log_path().parent / "runsdb"
DB_NAME = "rundb.csv"


def read_range(x, min_x, max_x):
    if x > 0:
        return x
    if min_x < 0 and max_x < 0:
        return min_x
    if min_x == max_x:
        return min_x
    return (min_x, max_x)


def save_run_info(
    name: str, log_path: Path, type_: str = "world", path: Path = DB_FOLDER
):
    try:
        path.mkdir(parents=True, exist_ok=True)
        with open(path / DB_NAME, "a") as f:
            f.write(f"{name},{type_},{str(log_path)}\n")
    except Exception as e:
        print(f"Failed to save run info: {str(e)}")


def default_tournament_path():
    """The default path to store tournament run info"""
    return default_log_path() / "tournaments"


def default_world_path():
    """The default path to store world run info"""
    return default_log_path() / "negotiations"


def print_progress(_, i, n) -> None:
    """Prints the progress of a tournament"""
    global n_completed, n_total
    n_completed = i + 1
    n_total = n
    print(
        f"{n_completed:04} of {n:04} worlds completed ({n_completed / n:0.2%})",
        flush=True,
    )


def shortest_unique_names(strs: List[str], sep="."):
    """
    Finds the shortest unique strings starting from the end of each input
    string based on the separator.

    The final strings will only be unique if the inputs are unique.

    Example:
        given ["a.b.c", "d.e.f", "a.d.c"] it will generate ["b.c", "f", "d.c"]
    """
    lsts = [_.split(sep) for _ in strs]
    names = [_[-1] for _ in lsts]
    if len(names) == len(set(names)):
        return names
    locs = defaultdict(list)
    for i, s in enumerate(names):
        locs[s].append(i)
    mapping = {"": ""}
    for s, l in locs.items():
        if len(s) < 1:
            continue
        if len(l) == 1:
            mapping[strs[l[0]]] = s
            continue
        strs_new = [sep.join(lsts[_][:-1]) for _ in l]
        prefixes = shortest_unique_names(strs_new, sep)
        for loc, prefix in zip(l, prefixes):
            x = sep.join([prefix, s])
            if x.startswith(sep):
                x = x[len(sep) :]
            mapping[strs[loc]] = x
    return [mapping[_] for _ in strs]


def nCr(n, r):
    return math.factorial(n) / math.factorial(r) / math.factorial(n - r)


click.option = partial(click.option, show_default=True)


@click.group()
def main():
    pass


@main.command(help="Runs an ANL 2024 tournament")
@click.option(
    "--parallel/--serial",
    default=True,
    help="Run a parallel/serial tournament on a single machine",
)
@click.option(
    "--name",
    "-n",
    default="random",
    help='The name of the tournament. The special value "random" will result in a random name',
)
@click.option(
    "--outcomes",
    "-o",
    default=DEFAULT2024SETTINGS["n_outcomes"] if not isinstance(DEFAULT2024SETTINGS["n_outcomes"], Iterable) else -1,  # type: ignore # type: ignore
    type=int,
    help="Number of outcomes in every scenario. If negative or zero, --min-outcomes and --max-outcomes will beused",
)
@click.option(
    "--min-outcomes",
    default=DEFAULT2024SETTINGS["n_outcomes"][0] if isinstance(DEFAULT2024SETTINGS["n_outcomes"], Iterable) else -1,  # type: ignore # type: ignore
    type=int,
    help="Minimum number of outcomes in every scenario. Only used of --outcomes is zero or negative",
)
@click.option(
    "--max-outcomes",
    default=DEFAULT2024SETTINGS["n_outcomes"][1] if isinstance(DEFAULT2024SETTINGS["n_outcomes"], Iterable) else -1,  # type: ignore # type: ignore
    type=int,
    help="Max number of outcomes in every scenario. Only used of --outcomes is zero or negative",
)
@click.option(
    "--scenarios",
    "-S",
    default=min(5, DEFAULT2024SETTINGS["n_scenarios"]),  # type: ignore
    type=int,
    help="Number of scenarios to generate",
)
@click.option(
    "--repetitions",
    "-r",
    default=DEFAULT2024SETTINGS["n_repetitions"],  # type: ignore
    type=int,
    help="Number of repetition for each negotiation",
)
@click.option(
    "--competitors",
    default=";".join(_().type_name.split(".")[-1] for _ in DEFAULT2024SETTINGS["competitors"]),  # type: ignore # type: ignore
    help="A semicolon (;) separated list of agent types to use for the competition. You"
    " can also pass the special value default for the default builtin"
    " agents",
)
@click.option(
    "--params",
    default="",
    help="A string encoding the parameters to pass for negotiators. Default is None. The format of this string is:\n"
    "'competitor:key:value,key:value,..key:value;competitor:key:value,key:value,...key:value\n"
    "For example:\nRVFitter:enable_logging:True:e:4.0;NashSeeker:nash_factor:1.0\nmeans: RVFitter's params are dict(enable_logging=True, e=True)"
    " and NashSeeker's params are dict(nash_factor=1.0)\n\nBe sure to use the python-correct format for values. For example put strings in single quots.",
)
@click.option(
    "--metric",
    default=DEFAULT2024SETTINGS["final_score"][0],  # type: ignore # type: ignore
    help="The metric to use for evaluating agents. Can be one of: utility, advantage, partner_welfare, welfare",
)
@click.option(
    "--stat",
    default=DEFAULT2024SETTINGS["final_score"][1],  # type: ignore # type: ignore
    help="The statistic applied to the metric to evaluate agents. Can be one of: mean, median, std, min, max",
)
@click.option(
    "--small/--normal",
    default=False,
    help="If --small, then a small tournament will be run and most parameters will be ignored",
)
@click.option(
    "--known-partner/--unknown-partner",
    default=DEFAULT2024SETTINGS["known_partner"],  # type: ignore
    help="Can the agent know its partner type?",
)
@click.option(
    "--save-logs/--no-logs",
    default=True,
    help="Whether to save logs.",
)
@click.option(
    "--timelimit",
    "-t",
    default=DEFAULT2024SETTINGS["time_limit"]  # type: ignore
    if not isinstance(DEFAULT2024SETTINGS["time_limit"], Iterable)  # type: ignore
    else -1,
    type=float,
    help="Number of seconds allowed in every negotiation. Negative numbers mean no-limit",
)
@click.option(
    "--min-timelimit",
    default=DEFAULT2024SETTINGS["time_limit"][0]  # type: ignore
    if isinstance(DEFAULT2024SETTINGS["time_limit"], Iterable)  # type: ignore
    else -1,
    type=int,
    help="Minimum number of seconds in every scenario. Only used of --timelimit is zero or negative",
)
@click.option(
    "--max-timelimit",
    default=DEFAULT2024SETTINGS["time_limit"][1]  # type: ignore
    if isinstance(DEFAULT2024SETTINGS["time_limit"], Iterable)  # type: ignore
    else -1,
    type=int,
    help="Max number of seconds in every scenario. Only used of --timelimit is zero or negative",
)
@click.option(
    "--steps",
    "-s",
    default=DEFAULT2024SETTINGS["n_steps"]  # type: ignore
    if not isinstance(DEFAULT2024SETTINGS["n_steps"], Iterable)  # type: ignore
    else -1,
    type=int,
    help="Number of negotiation rounds allowed in every negotiation. Negative numbers mean no-limit",
)
@click.option(
    "--min-steps",
    default=DEFAULT2024SETTINGS["n_steps"][0]  # type: ignore
    if isinstance(DEFAULT2024SETTINGS["n_steps"], Iterable)  # type: ignore
    else -1,
    type=int,
    help="Minimum number of steps in every scenario. Only used of --steps is zero or negative",
)
@click.option(
    "--max-steps",
    default=DEFAULT2024SETTINGS["n_steps"][1]  # type: ignore
    if isinstance(DEFAULT2024SETTINGS["n_steps"], Iterable)  # type: ignore
    else -1,
    type=int,
    help="Max number of steps in every scenario. Only used of --steps is zero or negative",
)
@click.option(
    "--pend",
    default=DEFAULT2024SETTINGS["pend"]  # type: ignore
    if not isinstance(DEFAULT2024SETTINGS["pend"], Iterable)  # type: ignore
    else -1,
    type=float,
    help="Probability of ending the negotiation at every round",
)
@click.option(
    "--min-pend",
    default=DEFAULT2024SETTINGS["pend"][0]  # type: ignore
    if isinstance(DEFAULT2024SETTINGS["pend"], Iterable)  # type: ignore
    else -1,
    help="Minimum pend in every scenario. Only used of --pend is zero or negative",
)
@click.option(
    "--max-pend",
    default=DEFAULT2024SETTINGS["pend"][1]  # type: ignore
    if isinstance(DEFAULT2024SETTINGS["pend"], Iterable)  # type: ignore
    else -1,
    help="Max pend in every scenario. Only used of --pend is zero or negative",
)
@click.option(
    "--self-play/--no-self-play",
    default=DEFAULT2024SETTINGS["self_play"],  # type: ignore
    type=bool,
    help="Allow self-play during the tournament",
)
@click.option(
    "--plot",
    "-p",
    default=0.1 if sys.platform in ("darwin", "linux") else 0,
    type=float,
    help="Fraction of negotiations to plot and save",
)
@click.option(
    "--verbosity",
    "-v",
    default=1,
    type=int,
    help="verbosity level",
)
@click.option(
    "--save_every",
    "-e",
    default=1,
    type=int,
    help="Number of negotiations after which to save stats",
)
@click.option(
    "--generator",
    default=DEFAULT2024SETTINGS["scenario_generator"],  # type: ignore
    type=str,
    help="The method to generate scenarios. Default is mix which generates a mix of scenario types containing zero-sum, monotonic and general scenarios",
)
@click.option(
    "--zerosum",
    "-z",
    default=DEFAULT2024SETTINGS["generator_params"].get("zerosum_fraction", 0.05),  # type: ignore
    type=float,
    help="Fraction of zero-sum scenarios (used when generator=mix)",
)
@click.option(
    "--monotonic",
    "-m",
    default=DEFAULT2024SETTINGS["generator_params"].get("monotonic_fraction", 0.25),  # type: ignore
    type=float,
    help="Fraction of monotonic scenarios (used when generator=mix)",
)
@click.option(
    "--curve",
    "-c",
    default=DEFAULT2024SETTINGS["generator_params"].get("curve_fraction", 0.5),  # type: ignore
    type=float,
    help="Fraction of monotonic and general scenarios generated using a Pareto curve not piecewise linear Pareto (used when generator=mix)",
)
@click.option(
    "--rotate/--no-rotate",
    default=DEFAULT2024SETTINGS["rotate_ufuns"],  # type: ignore
    help="Rotate utility functions when creating scenarios for the tournament",
)
@click.option(
    "--randomize/--no-randomize",
    default=DEFAULT2024SETTINGS["randomize_runs"],  # type: ignore
    help="Randomize the order of negotiations or not",
)
@click.option(
    "--two/--cartesian",
    default=False,  # type: ignore
    help="If --two is passed, a single negotiation will be conducted between the first two in the competitors list."
    " This is equivalent to passing --scenarios=1, --no-rotate --competitors=A;B --repetitions=1."
    " If only one competitor is given it is run against itself (with --self-platy assumed).",
)
@click.option(
    "--raise-exceptions/--ignore-exceptions",
    default=True,
    help="Whether to ignore agent exceptions",
)
@click.option(
    "--scenarios-path",
    default=None,
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    help="Path containing folders each representing a negotiation scenario.",
)
@click.option(
    "--settings-file",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="Path to a yaml/json file containing parameters to pass to the generator",
)
@click.option(
    "--path",
    default="",
    help="A path to be added to PYTHONPATH in which all competitors are stored. You can pass a : separated list of "
    "paths on linux/mac and a ; separated list in windows",
)
@click_config_file.configuration_option()
def tournament2024(
    parallel,
    generator,
    name,
    repetitions,
    raise_exceptions,
    competitors,
    params,
    verbosity,
    path,
    metric,
    save_logs,
    stat,
    known_partner,
    save_every,
    randomize,
    self_play,
    plot,
    pend,
    min_pend,
    max_pend,
    timelimit,
    min_timelimit,
    max_timelimit,
    steps,
    min_steps,
    max_steps,
    rotate,
    outcomes,
    min_outcomes,
    max_outcomes,
    scenarios,
    small,
    settings_file,
    zerosum,
    monotonic,
    curve,
    two,
    scenarios_path,
):
    if two:
        competitorslst = competitors.split(";")
        scenarios = 1
        rotate = False
        repetitions = 1
        parallel = False
        if len(competitorslst) == 1:
            competitorslst = [competitorslst[0], competitorslst[0]]
            self_play = True
        else:
            competitorslst = competitorslst[:2]
            self_play = False
        competitors = ";".join(competitorslst)
    generator_params = dict()
    # read settings of the scenario generator from the settings file if available
    if settings_file:
        generator_params = load(settings_file)
    # override settings with the appropriate ones based on the generator
    if generator == "mix":
        generator_params.update(
            zerosum_fraction=zerosum,
            monotonic_fraction=monotonic,
            curve_fraction=curve,
        )
    if small:
        scenarios = min(scenarios, 2)
        steps = -1
        min_steps, max_steps = (10, 1000)
        outcomes = 100
        repetitions = 1
        competitors = ";".join(competitors.split(";")[:3])

    steps = read_range(steps, min_steps, max_steps)
    outcomes = read_range(outcomes, min_outcomes, max_outcomes)
    timelimit = read_range(timelimit, min_timelimit, max_timelimit)
    pend = read_range(pend, min_pend, max_pend)
    if len(path) > 0:
        sys.path.append(path)
    if name == "random":
        name = unique_name(base="", rand_digits=0)

    all_competitors = competitors.split(";")
    all_params = [dict() for _ in all_competitors]
    if params:
        paramslst: list[str] = params.split(";")

        def parse_params(s: str, k: str) -> dict:
            vals = s.split(":")
            if len(vals) % 2 == 1:
                print(
                    f"[red]ERROR[red]: failed to parse {s} (the paramters of {k})"
                    f" in given competitor params {params} because it has an odd "
                    f"number of ':' separated parts ({len(vals)=}, {len(vals) % 2=}, {vals=})."
                )
                exit(1)
            d = dict()
            for i in range(0, len(vals), 2):
                d[vals[i]] = eval(vals[i + 1])
            return d

        params_map = defaultdict(dict)
        for s in paramslst:
            k, _, v = s.partition(":")
            params_map[k] = parse_params(v, k)

        for i, c in enumerate(all_competitors):
            all_params[i] = params_map.get(c, dict())

    def find_type_name(stem: str):
        for pre in (
            "",
            "anl.anl2024.negotiators.",
            "negmas.sao.negotiators.",
            "negmas.genius.gnegotiators.",
        ):
            s = pre + stem
            try:
                get_class(s)
                return s
            except:
                pass
        print(f"[red]ERROR[/red] Unknown Competitor Type: {stem}")
        sys.exit()

    for i, cp in enumerate(all_competitors):
        all_competitors[i] = find_type_name(cp)  # type: ignore

    if not all_competitors:
        all_competitors = DEFAULT_AN2024_COMPETITORS

    print(f"Tournament will be run between {len(all_competitors)} agents: ")
    print(all_competitors)
    n_params = sum(len(_) for _ in all_params)
    if n_params > 0:
        print(f"Will use the following competitor parameters:")
        for c, p in zip(all_competitors, all_params):
            if not p:
                continue
            print(f"{c}:\n\t{p}")
    if not isinstance(steps, tuple) and steps <= 0:
        steps = None
    if not isinstance(timelimit, tuple) and timelimit <= 0:
        timelimit = None
    if (
        steps is None
        and timelimit is None
        and not isinstance(pend, tuple)
        and pend <= 0.0
    ):
        print(
            f"[red]ERROR[/red] You specified no way to end the negotiation. You MUST pass either --steps, --timelimit or --pend (or the --min, --max versions of them)"
        )
        sys.exit(1)
    loaded_scenarios = []
    if scenarios:
        print(f"Will generate {scenarios} scenarios of {outcomes} outcomes each.")
    if scenarios_path is not None:
        for path in chain([scenarios_path], Path(scenarios_path).glob("**/*")):
            path = Path(path)
            if not path.is_dir():
                continue
            if not Scenario.is_loadable(path):
                print(f"{path} is not loadable")
                continue
            try:
                scenario = Scenario.load(path, safe_parsing=False)
                if scenario:
                    loaded_scenarios.append(scenario)
            except Exception as e:
                warnings.warn(f"Could not load scenario from {path}. Error: {str(e)}")

        print(
            f"Will use {len(loaded_scenarios)} scenarios loaded from {scenarios_path}."
        )
    print(f"Negotiations will end if any of the following conditions is satisfied:")
    if steps is not None:
        print(f"\tN. Rounds: {steps}")
    if timelimit is not None:
        print(f"\tN. Seconds: {timelimit}")
    if pend is not None and (isinstance(pend, tuple) or pend > 0):
        print(f"\tProbability of ending each round : {pend}")
    if len(loaded_scenarios) + scenarios == 0:
        print(
            f"You must either pass --scenarios with the number of scenarios to be generated or pass --scenarios-path with a folder containing scenarios to use.\nYou are passing {scenarios=}, {scenarios_path=}. \nWill exit"
        )
        exit()
    tic = perf_counter()
    results = anl2024_tournament(
        scenarios=loaded_scenarios,
        n_scenarios=scenarios,
        n_outcomes=outcomes,
        competitors=all_competitors,  # type: ignore
        competitor_params=all_params,
        rotate_ufuns=rotate,
        n_repetitions=repetitions,
        n_steps=steps,
        time_limit=timelimit,
        pend=pend,
        name=name,
        nologs=not save_logs,
        njobs=0 if parallel else -1,
        plot_fraction=plot,
        verbosity=verbosity,
        self_play=self_play,
        randomize_runs=randomize,
        save_every=save_every,
        known_partner=known_partner,
        final_score=(metric, stat),
        scenario_generator=generator,
        generator_params=generator_params,
        raise_exceptions=raise_exceptions,
    )
    if verbosity <= 0:
        print(results.final_scores)
    print(f"Done in {humanize_time(perf_counter() - tic, show_ms=True)}")
    if save_logs:
        print(f"Detailed logs are stored at: {DEFAULT_TOURNAMENT_PATH / name}")


@main.command(
    help="Generates sceanrios a la ANL.\n\nYou must pass the path to save the scenarios into as the first parameter"
)
@click.argument(
    "path",
    type=click.Path(file_okay=False),
)
@click.option(
    "--outcomes",
    "-o",
    default=DEFAULT2024SETTINGS["n_outcomes"] if not isinstance(DEFAULT2024SETTINGS["n_outcomes"], Iterable) else -1,  # type: ignore # type: ignore
    type=int,
    help="Number of outcomes in every scenario. If negative or zero, --min-outcomes and --max-outcomes will beused",
)
@click.option(
    "--min-outcomes",
    default=DEFAULT2024SETTINGS["n_outcomes"][0] if isinstance(DEFAULT2024SETTINGS["n_outcomes"], Iterable) else -1,  # type: ignore # type: ignore
    type=int,
    help="Minimum number of outcomes in every scenario. Only used of --outcomes is zero or negative",
)
@click.option(
    "--max-outcomes",
    default=DEFAULT2024SETTINGS["n_outcomes"][1] if isinstance(DEFAULT2024SETTINGS["n_outcomes"], Iterable) else -1,  # type: ignore # type: ignore
    type=int,
    help="Max number of outcomes in every scenario. Only used of --outcomes is zero or negative",
)
@click.option(
    "--scenarios",
    "-S",
    default=min(5, DEFAULT2024SETTINGS["n_scenarios"]),  # type: ignore
    type=int,
    help="Number of scenarios to generate",
)
@click.option(
    "--generator",
    default=DEFAULT2024SETTINGS["scenario_generator"],  # type: ignore
    type=str,
    help="The method to generate scenarios. Default is mix which generates a mix of scenario types containing zero-sum, monotonic and general scenarios",
)
@click.option(
    "--zerosum",
    "-z",
    default=DEFAULT2024SETTINGS["generator_params"].get("zerosum_fraction", 0.05),  # type: ignore
    type=float,
    help="Fraction of zero-sum scenarios (used when generator=mix)",
)
@click.option(
    "--monotonic",
    "-m",
    default=DEFAULT2024SETTINGS["generator_params"].get("monotonic_fraction", 0.25),  # type: ignore
    type=float,
    help="Fraction of monotonic scenarios (used when generator=mix)",
)
@click.option(
    "--curve",
    "-c",
    default=DEFAULT2024SETTINGS["generator_params"].get("curve_fraction", 0.5),  # type: ignore
    type=float,
    help="Fraction of monotonic and general scenarios generated using a Pareto curve not piecewise linear Pareto (used when generator=mix)",
)
@click.option(
    "--settings-file",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="Path to a yaml/json file containing parameters to pass to the generator",
)
@click.option(
    "--plot",
    "-p",
    default=0.1 if sys.platform in ("darwin", "linux") else 0,
    type=float,
    help="Fraction of negotiations to plot and save",
)
@click_config_file.configuration_option()
def make_scenarios(
    path,
    scenarios,
    outcomes,
    min_outcomes,
    max_outcomes,
    generator,
    settings_file,
    zerosum,
    monotonic,
    curve,
    plot,
):
    if scenarios == 0:
        print(f"You must pass --scenarios with the number of scenarios to be generated")
        exit()
    generator_params = dict()
    # read settings of the scenario generator from the settings file if available
    if settings_file:
        generator_params = load(settings_file)
    # override settings with the appropriate ones based on the generator
    if generator == "mix":
        generator_params.update(
            zerosum_fraction=zerosum,
            monotonic_fraction=monotonic,
            curve_fraction=curve,
        )

    outcomes = read_range(outcomes, min_outcomes, max_outcomes)

    print(f"Will generate {scenarios} scenarios of {outcomes} outcomes each.")
    scenario_generator = GENERAROR_MAP[generator]
    scenarios = scenario_generator(
        n_scenarios=scenarios, n_outcomes=outcomes, **generator_params
    )
    path = Path(path)
    for s in scenarios:
        mypath = path / s.outcome_space.name  # type: ignore
        s.dumpas(mypath)  # type: ignore
        if plot:
            plot_offline_run(
                trace=[],
                ids=["First", "Second"],
                ufuns=s.ufuns,  # type: ignore
                agreement=None,
                timedout=False,
                broken=False,
                has_error=False,
                names=["First", "Second"],
                save_fig=True,
                path=mypath,
                fig_name="fig.png",
                only2d=True,
                show_annotations=False,
                show_agreement=False,
                show_pareto_distance=False,
                show_nash_distance=False,
                show_kalai_distance=False,
                show_max_welfare_distance=False,
                show_max_relative_welfare_distance=False,
                show_end_reason=False,
                show_reserved=True,
                show_total_time=False,
                show_relative_time=False,
                show_n_steps=False,
            )
            plt.close()


@main.command(help="Displays ANL and NegMAS versions")
def version():
    print(f"anl: {anl.__version__} (NegMAS: {negmas.__version__})")


if __name__ == "__main__":
    main()



import random

from negmas import nash_points, pareto_frontier
from negmas.outcomes import Outcome
from negmas.sao import ResponseType, SAONegotiator, SAOResponse, SAOState

__all__ = ["NashSeeker"]


class NashSeeker(SAONegotiator):
    """Assumes that the opponent has a fixed reserved value and seeks the Nash equilibrium.

    Args:
        opponent_reserved_value: Assumed reserved value for the opponent
        nash_factor: Fraction (or multiple) of the agent utility at the Nash Point (assuming the `opponent_reserved_value`) that is acceptable

    """

    def __init__(
        self, *args, opponent_reserved_value: float = 0.25, nash_factor=0.9, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._opponent_r = opponent_reserved_value
        self._outcomes: list[Outcome] = []
        self._min_acceptable = float("inf")
        self._nash_factor = nash_factor
        self._best: Outcome = None  # type: ignore

    def on_preferences_changed(self, changes):
        _ = changes  # silenting a typing warning
        # This callback is called at the start of the negotiation after the ufun is set
        assert self.ufun is not None and self.ufun.outcome_space is not None
        # save my best outcome for later use
        self._best = self.ufun.best()
        # check that I have access to  the opponent ufun
        assert self.opponent_ufun is not None
        # set the reserved value of the opponent
        self.opponent_ufun.reserved_value = self._opponent_r
        # consider my and my parther's ufuns
        ufuns = (self.ufun, self.opponent_ufun)
        # list all outcomes
        outcomes = list(self.ufun.outcome_space.enumerate_or_sample())
        # find the pareto-front and the nash point
        frontier_utils, frontier_indices = pareto_frontier(ufuns, outcomes)
        frontier_outcomes = [outcomes[_] for _ in frontier_indices]
        my_frontier_utils = [_[0] for _ in frontier_utils]
        nash = nash_points(ufuns, frontier_utils)  # type: ignore
        if nash:
            # find my utility at the Nash Bargaining Solution.
            my_nash_utility = nash[0][0][0]
        else:
            my_nash_utility = 0.5 * (float(self.ufun.max()) + self.ufun.reserved_value)
        # Set the acceptable utility limit
        self._min_acceptable = my_nash_utility * self._nash_factor
        # Set the set of outcomes to offer from
        self._outcomes = [
            w
            for u, w in zip(my_frontier_utils, frontier_outcomes)
            if u >= self._min_acceptable
        ]

    def __call__(self, state: SAOState) -> SAOResponse:
        # just assert that I have a ufun and I know the outcome space.
        assert self.ufun is not None and self.ufun.outcome_space is not None
        # read the current offer from the state. None means I am starting the negotiation
        offer = state.current_offer
        # Accept the offer if its utility is higher than my utility at the Nash Bargaining Solution with the assumed opponent ufun
        if offer and float(self.ufun(offer)) >= self._min_acceptable:
            return SAOResponse(ResponseType.ACCEPT_OFFER, offer)
        # If I could not find the Nash Bargaining Solution, just offer my best outcome forever.
        if not self._outcomes:
            return SAOResponse(ResponseType.REJECT_OFFER, self._best)
        # Offer some outcome with high utility relative to the Nash Bargaining Solution
        return SAOResponse(ResponseType.REJECT_OFFER, random.choice(self._outcomes))



import random

import numpy as np
from negmas import Outcome, ResponseType, SAONegotiator, SAOResponse, SAOState
from scipy.optimize import curve_fit

__all__ = ["RVFitter"]


def aspiration_function(t, mx, rv, e):
    return (mx - rv) * (1.0 - np.power(t, e)) + rv


class RVFitter(SAONegotiator):
    """A simple negotiator that uses curve fitting to learn the reserved value.

    Args:
        min_unique_utilities: Number of different offers from the opponent before starting to
                              attempt learning their reserved value.
        e: The concession exponent used for the agent's offering strategy
        stochasticity: The level of stochasticity in the offers.
        enable_logging: If given, a log will be stored  for the estimates.

    Remarks:

        - Assumes that the opponent is using a time-based offering strategy that offers
          the outcome at utility $u(t) = (u_0 - r) - r \\exp(t^e)$ where $u_0$ is the utility of
          the first offer (directly read from the opponent ufun), $e$ is an exponent that controls the
          concession rate and $r$ is the reserved value we want to learn.
        - After it receives offers with enough different utilities, it starts finding the optimal values
          for $e$ and $r$.
        - When it is time to respond, RVFitter, calculates the set of rational outcomes **for both agents**
          based on its knowledge of the opponent ufun (given) and reserved value (learned). It then applies
          the same concession curve defined above to concede over an ordered list of these outcomes.
        - Is this better than using the same concession curve on the outcome space without even trying to learn
          the opponent reserved value? Maybe sometimes but empirical evaluation shows that it is not in general.
        - Note that the way we check for availability of enough data for training is based on the uniqueness of
          the utility of offers from the opponent (for the opponent). Given that these are real values, this approach
          is suspect because of rounding errors. If two outcomes have the same utility they may appear to have different
          but very close utilities because or rounding errors (or genuine very small differences). Such differences should
          be ignored.
        - Note also that we start assuming that the opponent reserved value is 0.0 which means that we are only restricted
          with our own reserved values when calculating the rational outcomes. This is the best case scenario for us because
          we have MORE negotiation power when the partner has LOWER utility.
    """

    def __init__(
        self,
        *args,
        min_unique_utilities: int = 10,
        e: float = 5.0,
        stochasticity: float = 0.1,
        enable_logging: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.min_unique_utilities = min_unique_utilities
        self.e = e
        self.stochasticity = stochasticity
        # keeps track of times at which the opponent offers
        self.opponent_times: list[float] = []
        # keeps track of opponent utilities of its offers
        self.opponent_utilities: list[float] = []
        # keeps track of the our last estimate of the opponent reserved value
        self._past_oppnent_rv = 0.0
        # keeps track of the rational outcome set given our estimate of the
        # opponent reserved value and our knowledge of ours
        self._rational: list[tuple[float, float, Outcome]] = []
        self._enable_logging = enable_logging

    def __call__(self, state: SAOState) -> SAOResponse:
        assert self.ufun and self.opponent_ufun
        # update the opponent reserved value in self.opponent_ufun
        self.update_reserved_value(state)
        # rune the acceptance strategy and if the offer received is acceptable, accept it
        if self.is_acceptable(state):
            return SAOResponse(ResponseType.ACCEPT_OFFER, state.current_offer)
        # The offering strategy
        # We only update our estimate of the rational list of outcomes if it is not set or
        # there is a change in estimated reserved value
        if (
            not self._rational
            or abs(self.opponent_ufun.reserved_value - self._past_oppnent_rv) > 1e-3
        ):
            # The rational set of outcomes sorted dependingly according to our utility function
            # and the opponent utility function (in that order).
            self._rational = sorted(
                [
                    (my_util, opp_util, _)
                    for _ in self.nmi.outcome_space.enumerate_or_sample(
                        levels=10, max_cardinality=100_000
                    )
                    if (my_util := float(self.ufun(_))) > self.ufun.reserved_value
                    and (opp_util := float(self.opponent_ufun(_)))
                    > self.opponent_ufun.reserved_value
                ],
            )
        # If there are no rational outcomes (i.e. our estimate of the opponent rv is very wrogn),
        # then just revert to offering our top offer
        if not self._rational:
            return SAOResponse(ResponseType.REJECT_OFFER, self.ufun.best())
        # find our aspiration level (value between 0 and 1) the higher the higher utility we require
        asp = aspiration_function(state.relative_time, 1.0, 0.0, self.e)
        # find the index of the rational outcome at the aspiration level (in the rational set of outcomes)
        n_rational = len(self._rational)
        max_rational = n_rational - 1
        min_indx = max(0, min(max_rational, int(asp * max_rational)))
        # find current stochasticity which goes down from the set level to zero linearly
        s = aspiration_function(state.relative_time, self.stochasticity, 0.0, 1.0)
        # find the index of the maximum utility we require based on stochasticity (going down over time)
        max_indx = max(0, min(int(min_indx + s * n_rational), max_rational))
        # offer an outcome in the selected range
        indx = random.randint(min_indx, max_indx) if min_indx != max_indx else min_indx
        outcome = self._rational[indx][-1]
        return SAOResponse(ResponseType.REJECT_OFFER, outcome)

    def is_acceptable(self, state: SAOState) -> bool:
        # The acceptance strategy
        assert self.ufun and self.opponent_ufun
        # get the offer from the mechanism state
        offer = state.current_offer
        # If there is no offer, there is nothing to accept
        if offer is None:
            return False
        # Find the current aspiration level
        asp = aspiration_function(
            state.relative_time, 1.0, self.ufun.reserved_value, self.e
        )
        # accept if the utility of the received offer is higher than
        # the current aspiration
        return float(self.ufun(offer)) >= asp

    def update_reserved_value(self, state: SAOState):
        # Learns the reserved value of the partner
        assert self.opponent_ufun is not None
        # extract the current offer from the state
        offer = state.current_offer
        if offer is None:
            return
        # save to the list of utilities received from the opponent and their times
        self.opponent_utilities.append(float(self.opponent_ufun(offer)))
        self.opponent_times.append(state.relative_time)

        # If we do not have enough data, just assume that the opponent
        # reserved value is zero
        n_unique = len(set(self.opponent_utilities))
        if n_unique < self.min_unique_utilities:
            self._past_oppnent_rv = 0.0
            self.opponent_ufun.reserved_value = 0.0
            return
        # Use curve fitting to estimate the opponent reserved value
        # We assume the following:
        # - The opponent is using a concession strategy with an exponent between 0.2, 5.0
        # - The opponent never offers outcomes lower than their reserved value which means
        #   that their rv must be no higher than the worst outcome they offered for themselves.
        bounds = ((0.2, 0.0), (5.0, min(self.opponent_utilities)))
        err = ""
        try:
            optimal_vals, _ = curve_fit(
                lambda x, e, rv: aspiration_function(
                    x, self.opponent_utilities[0], rv, e
                ),
                self.opponent_times,
                self.opponent_utilities,
                bounds=bounds,
            )
            self._past_oppnent_rv = self.opponent_ufun.reserved_value
            self.opponent_ufun.reserved_value = optimal_vals[1]
        except Exception as e:
            err, optimal_vals = f"{str(e)}", [None, None]

        # log my estimate
        if self._enable_logging:
            self.nmi.log_info(
                self.id,
                dict(
                    estimated_rv=self.opponent_ufun.reserved_value,
                    n_unique=n_unique,
                    opponent_utility=self.opponent_utilities[-1],
                    estimated_exponent=optimal_vals[0],
                    estimated_max=self.opponent_utilities[0],
                    error=err,
                ),
            )



import random

from negmas.outcomes import Outcome
from negmas.preferences import PresortingInverseUtilityFunction
from negmas.sao import ResponseType, SAONegotiator, SAOResponse, SAOState

__all__ = ["MiCRO"]


class MiCRO(SAONegotiator):
    """
    A simple implementation of the MiCRO negotiation strategy

    Remarks:
        - This is a simplified implementation of the MiCRO strategy.
        - It is not guaranteed to exactly match the published work.
        - MiCRO was introduced here:
          de Jonge, Dave. "An Analysis of the Linear Bilateral ANAC Domains Using the MiCRO Benchmark Strategy."
          Proceedings of the Thirty-First International Joint Conference on Artificial Intelligence, IJCAI. 2022.
        - Note that MiCRO works optimally if both negotiators can concede all the way to agreement. If one of them
          has a high reservation value preventing it from doing so, or if the allowed number of steps is small, MiCRO
          will not reach agreement (even against itself).
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # initialize local variables
        self.worst_offer_utility: float = float("inf")
        self.sorter = None
        self._received, self._sent = set(), set()

    def __call__(self, state: SAOState) -> SAOResponse:
        # The main implementation of the MiCRO strategy
        assert self.ufun
        # initialize the sorter (This should better be done in on_negotiation_start() to allow for reuse but this is not needed in ANL)
        if self.sorter is None:
            # A sorter, sorts a ufun and can be used to get outcomes using their utiility
            self.sorter = PresortingInverseUtilityFunction(
                self.ufun, rational_only=True, eps=-1, rel_eps=-1
            )
            # Initialize the sorter. This is an O(nlog n) operation where n is the number of outcomes
            self.sorter.init()
        # get the current offer and prepare for rejecting it
        offer = state.current_offer

        # If I received something, check for acceptance
        if offer is not None:
            self._received.add(offer)

        # Find out my next offer and the acceptable offer
        will_concede = len(self._sent) <= len(self._received)
        # My next offer is either a conceding outcome if will_concede or sampled randomly from my past offers
        next_offer = (
            self.sample_sent() if not will_concede else self.sorter.next_worse()
        )
        # If I exhausted all my rational offers, do not concede
        if next_offer is None:
            will_concede, next_offer = False, self.sample_sent()
        else:
            next_utility = float(self.ufun(next_offer))
            if next_utility < self.ufun.reserved_value:
                will_concede, next_offer = False, self.sample_sent()
        next_utility = float(self.ufun(next_offer))
        # Find my acceptable outcome. It will None if I did not offer anything yet.
        acceptable_utility = (
            self.worst_offer_utility if not will_concede else next_utility
        )

        # The Acceptance Policy of MiCRO
        # accept if the offer is not worse than my acceptable offer if I am conceding or the best so far if I am not
        offer_utility = float(self.ufun(offer))
        if (
            offer is not None
            and offer_utility >= acceptable_utility
            and offer_utility >= self.ufun.reserved_value
        ):
            return SAOResponse(ResponseType.ACCEPT_OFFER, offer)
        # If I cannot find any offers, I know that there are NO rational outcomes in this negotiation for me and will end it.
        if next_offer is None:
            return SAOResponse(ResponseType.END_NEGOTIATION, None)
        # Offer my next-offer and record it
        self._sent.add(next_offer)
        self.worst_offer_utility = next_utility
        return SAOResponse(ResponseType.REJECT_OFFER, next_offer)

    def sample_sent(self) -> Outcome | None:
        # Get an outcome from the set I sent so far (or my best if I sent nothing)
        if not len(self._sent):
            return None
        return random.choice(list(self._sent))



import os

__all__ = [
    "ANL_RUNALL_TESTS",
    "ANL_FASTRUN",
    "ANL_RUN_TUTORIAL2",
    "ANL_RUN_GENIUS",
    "ANL_RUN_TOURNAMENTS",
    "ANL_RUN_TEMP_FAILING",
]


def is_enabled(val: str) -> bool:
    return os.environ.get(val, "").lower() in ("true", "yes")


def isnot_disabled(val: str) -> bool:
    return os.environ.get(val, "").lower() not in ("false", "no")


ANL_RUNALL_TESTS = is_enabled("ANL_RUNALL_TESTS")
ANL_ON_GITHUB = is_enabled("GITHUB_ACTIONS")
ANL_FASTRUN = (is_enabled("ANL_FASTRUN") or ANL_ON_GITHUB) and not ANL_RUNALL_TESTS
ANL_RUN_TEMP_FAILING = is_enabled("ANL_RUN_TEMP_FAILING")
ANL_RUN_GENIUS = is_enabled("ANL_RUN_GENIUS")
ANL_RUN_TOURNAMENTS = is_enabled("ANL_RUN_TOURNAMENTS")
ANL_RUN_TUTORIAL2 = isnot_disabled("ANL_RUN_TUTORIAL2")
ANL_RUN_NOTEBOOKS = isnot_disabled("ANL_RUN_NOTEBOOKS")



from negmas.sao.negotiators import (
    AspirationNegotiator,
    BoulwareTBNegotiator,
    ConcederTBNegotiator,
    LinearTBNegotiator,
    NaiveTitForTatNegotiator,
)

__all__ = ["Linear", "Conceder", "Boulware", "NaiveTitForTat"]


class NaiveTitForTat(NaiveTitForTatNegotiator):
    """
    A simple behavioral strategy that assumes a zero-sum game
    """


class StochasticLinear(AspirationNegotiator):
    """
    Time-based linear negotiation strategy (offers above the limit instead of at it)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            max_aspiration=1.0,
            aspiration_type="linear",
            tolerance=0.00001,
            stochastic=True,
            presort=True,
            **kwargs
        )


class StochasticConceder(AspirationNegotiator):
    """
    Time-based conceder negotiation strategy (offers above the limit instead of at it)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            max_aspiration=1.0,
            aspiration_type="conceder",
            tolerance=0.00001,
            stochastic=True,
            presort=True,
            **kwargs
        )


class StochasticBoulware(AspirationNegotiator):
    """
    Time-based boulware negotiation strategy (offers above the limit instead of at it)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            max_aspiration=1.0,
            aspiration_type="boulware",
            tolerance=0.00001,
            stochastic=True,
            presort=True,
            **kwargs
        )


class Linear(LinearTBNegotiator):
    """
    Time-based linear negotiation strategy
    """

    ...


class Conceder(ConcederTBNegotiator):
    """
    Time-based conceder negotiation strategy
    """


class Boulware(BoulwareTBNegotiator):
    """
    Time-based boulware negotiation strategy
    """
